@@ -590,7 +590,7 @@ impl Trigger {
 /// Represents the annotated given type for constants, input streams, etc.
 /// It is converted from the AST type and an input for the type checker.
 /// After typechecking HirType is used to represent all type information.
-#[derive(Debug, PartialEq, Eq, Clone)]
+#[derive(Debug, PartialEq, Eq, Clone, Hash)]
 pub(crate) enum AnnotatedType {
     Int(u32),
     Float(u32),
