@@ -6,11 +6,9 @@ pub(crate) mod types;
 
 use std::collections::HashMap;
 
-use rtlola_reporting::Handler;
+use rtlola_reporting::RtLolaError;
 
-use self::dependencies::{DependencyErr, DependencyGraph, Streamdependencies, Windowdependencies};
-use self::memory_bounds::MemBoundErr;
-use self::ordering::OrderErr;
+use self::dependencies::{DependencyGraph, Streamdependencies, Windowdependencies};
 use self::types::HirType;
 use crate::hir::{ExprId, Hir, SRef, WRef};
 use crate::modes::memory_bounds::MemorizationBound;
@@ -24,19 +22,16 @@ use crate::type_check::{ConcreteValueType, StreamType};
 /// Each mode implements a separate trait defining the functionality that is added by the new mode, e.g., the [TypedMode] implements the [TypedTrait], providing an interface to get the types of a stream or expression.
 /// With a new mode, a compiler flag derives the functionality of the previous modes.
 /// The [RtLolaHir](crate::RtLolaHir) progesses the following modes:
-/// [BaseMode] -> [DepAnaMode] -> [TypedMode] -> [OrderedMode] -> [MemBoundMode] -> [CompleteMode]
+/// [BaseMode] -> [TypedMode] -> [DepAnaMode] -> [OrderedMode] -> [MemBoundMode] -> [CompleteMode]
 pub trait HirMode {}
 
 /// Defines the functionality to progress one mode to the next one
 pub trait HirStage: Sized {
-    /// Defines the Error type of the `progress` function
-    type Error;
-
     /// Defines the next mode that is produced by the `progress` function
     type NextStage: HirMode;
 
     /// Returns an [RtLolaHir](crate::RtLolaHir) with additional functionality
-    fn progress(self, handler: &Handler) -> Result<Hir<Self::NextStage>, Self::Error>;
+    fn progress(self) -> Result<Hir<Self::NextStage>, RtLolaError>;
 }
 
 /// Represents the first stage in the [RtLolaHir](crate::RtLolaHir)(crate::RtLolaHir)
@@ -47,13 +42,12 @@ pub trait HirStage: Sized {
 pub struct BaseMode {}
 
 impl HirStage for Hir<BaseMode> {
-    type Error = DependencyErr;
-    type NextStage = DepAnaMode;
+    type NextStage = TypedMode;
 
-    fn progress(self, _handler: &Handler) -> Result<Hir<Self::NextStage>, Self::Error> {
-        let dependencies = DepAna::analyze(&self)?;
+    fn progress(self) -> Result<Hir<Self::NextStage>, RtLolaError> {
+        let tts = crate::type_check::type_check(&self)?;
 
-        let mode = DepAnaMode { dependencies };
+        let mode = TypedMode { types: tts };
 
         Ok(Hir {
             inputs: self.inputs,
@@ -68,124 +62,16 @@ impl HirStage for Hir<BaseMode> {
 }
 
 impl Hir<BaseMode> {
-    /// Returns the [RtLolaHir](crate::RtLolaHir) with additional information about the dependencies between streams
-    ///
-    /// The function returns the [RtLolaHir](crate::RtLolaHir) after the dependency analysis.
-    /// The new mode implements the same functionality as the [BaseMode] and additionally contains the dependencies between streams in the specification.
-    /// The function moves the information of the previous mode to the new one and therefore destroys the current mode.
-    ///
-    /// # Fails
-    /// The function returns a [DependencyErr] if the specification is not well-formed.
-    pub fn analyze_dependencies(self, handler: &Handler) -> Result<Hir<DepAnaMode>, DependencyErr> {
-        self.progress(handler)
-    }
-}
-
-/// Represents the results of the dependency analysis
-#[derive(Debug, Clone)]
-pub struct DepAna {
-    direct_accesses: Streamdependencies,
-    transitive_accesses: Streamdependencies,
-    direct_accessed_by: Streamdependencies,
-    transitive_accessed_by: Streamdependencies,
-    aggregated_by: Windowdependencies,
-    aggregates: Windowdependencies,
-    graph: DependencyGraph,
-}
-
-/// Represents the mode after the dependency analysis
-///
-/// This struct represents the mode after the dependency analysis.
-/// Besides this result, this mode has the same functionality as all the previous modes.
-/// The [DepAnaTrait] defines the new functionality of the mode.
-#[covers_functionality(DepAnaTrait, dependencies)]
-#[derive(Debug, Clone, HirMode)]
-pub struct DepAnaMode {
-    dependencies: DepAna,
-}
-
-/// Describes the functionality of a mode after analyzing the dependencies
-#[mode_functionality]
-pub trait DepAnaTrait {
-    /// Returns all streams that are direct accessed by `who`
-    ///
-    /// The function returns all streams that are direct accessed by `who`.
-    /// A stream `who` accesses a stream `res`, if the stream expression, the spawn condition and definition, the filter condition, or the close condition of 'who' has a stream or window lookup to `res`.
-    /// Direct accesses are all accesses appearing in the expressions of the stream itself.
-    fn direct_accesses(&self, who: SRef) -> Vec<SRef>;
-
-    /// Returns all streams that are transitive accessed by `who`
-    ///
-    /// The function returns all streams that are transitive accessed by `who`.
-    /// A stream `who` accesses a stream `res`, if the stream expression, the spawn condition and definition, the filter condition, or the close condition of 'who' has a stream or window lookup to 'res'.
-    /// Transitive accesses are all accesses appearing in the expressions of the stream itself or indirect by another stream lookup.
-    fn transitive_accesses(&self, who: SRef) -> Vec<SRef>;
-
-    /// Returns all streams that direct access `who`
-    ///
-    /// The function returns all streams that direct access `who`.
-    /// A stream `who` is accessed by a stream `res`, if the stream expression, the spawn condition and definition, the filter condition, or the close condition of 'res' has a stream or window lookup to 'who'.
-    /// Direct accesses are all accesses appearing in the expressions of the stream itself.
-    fn direct_accessed_by(&self, who: SRef) -> Vec<SRef>;
-
-    /// Returns all streams that transitive access `who`
-    ///
-    /// The function returns all streams that transitive access `who`.
-    /// A stream `who` is accessed by a stream `res`, if the stream expression, the spawn condition and definition, the filter condition, or the close condition of 'res' has a stream or window lookup to 'who'.
-    /// Transitive accesses are all accesses appearing in the expressions of the stream itself or indirect by another stream lookup.
-    fn transitive_accessed_by(&self, who: SRef) -> Vec<SRef>;
-
-    /// Returns all windows that aggregate `who` and the stream that uses the window
-    ///
-    /// The function returns all windows that aggregate `who` and the stream that uses the window.
-    /// The result contains only the windows that are direct.
-    fn aggregated_by(&self, who: SRef) -> Vec<(SRef, WRef)>; // (non-transitive)
-
-    /// Returns all windows that are used in `who` and the corresponding stream that is aggregated
-    ///
-    /// The function returns all windows that are used in `who` and the corresponding stream that is aggregated.
-    /// The result contains only the windows that are direct.
-    fn aggregates(&self, who: SRef) -> Vec<(SRef, WRef)>; // (non-transitive)
-
-    /// Returns the (Dependency Graph)[DependencyGraph] of the specification
-    fn graph(&self) -> &DependencyGraph;
-}
-
-impl HirStage for Hir<DepAnaMode> {
-    type Error = String;
-    type NextStage = TypedMode;
-
-    fn progress(self, handler: &Handler) -> Result<Hir<Self::NextStage>, Self::Error> {
-        let tts = crate::type_check::type_check(&self, handler)?;
-
-        let mode = TypedMode {
-            dependencies: self.mode.dependencies,
-            types: tts,
-        };
-
-        Ok(Hir {
-            inputs: self.inputs,
-            outputs: self.outputs,
-            triggers: self.triggers,
-            next_output_ref: self.next_output_ref,
-            next_input_ref: self.next_input_ref,
-            expr_maps: self.expr_maps,
-            mode,
-        })
-    }
-}
-
-impl Hir<DepAnaMode> {
     /// Returns the [RtLolaHir](crate::RtLolaHir) with the type information for each stream and expression
     ///
     /// The function returns the [RtLolaHir](crate::RtLolaHir) after the type analysis.
-    /// The new mode implements the same functionality as the [DepAnaMode] and additionally holds for each stream and expression its [StreamType].
+    /// The new mode implements the same functionality as the [BaseMode] and additionally holds for each stream and expression its [StreamType].
     /// The function moves the information of the previous mode to the new one and therefore destroys the current mode.
     ///
     /// # Fails
     /// The function fails if the type checker finds a type error in the specification and returns a string with a detailed description.
-    pub fn check_types(self, handler: &Handler) -> Result<Hir<TypedMode>, String> {
-        self.progress(handler)
+    pub fn check_types(self) -> Result<Hir<TypedMode>, RtLolaError> {
+        self.progress()
     }
 }
 
@@ -202,11 +88,9 @@ pub struct Typed {
 /// This struct represents the mode after the type checker call.
 /// Besides this result, this mode has the same functionality as all the previous modes.
 /// The [TypedTrait] defines the new functionality of the mode.
-#[covers_functionality(DepAnaTrait, dependencies)]
 #[covers_functionality(TypedTrait, types)]
 #[derive(Debug, Clone, HirMode)]
 pub struct TypedMode {
-    dependencies: DepAna,
     types: Typed,
 }
 
@@ -259,10 +143,118 @@ pub trait TypedTrait {
 }
 
 impl HirStage for Hir<TypedMode> {
-    type Error = OrderErr;
+    type NextStage = DepAnaMode;
+
+    fn progress(self) -> Result<Hir<Self::NextStage>, RtLolaError> {
+        let dependencies = DepAna::analyze(&self)?;
+
+        let mode = DepAnaMode {
+            dependencies,
+            types: self.mode.types,
+        };
+
+        Ok(Hir {
+            inputs: self.inputs,
+            outputs: self.outputs,
+            triggers: self.triggers,
+            next_output_ref: self.next_output_ref,
+            next_input_ref: self.next_input_ref,
+            expr_maps: self.expr_maps,
+            mode,
+        })
+    }
+}
+
+impl Hir<TypedMode> {
+    /// Returns the [RtLolaHir](crate::RtLolaHir) with additional information about the dependencies between streams
+    ///
+    /// The function returns the [RtLolaHir](crate::RtLolaHir) after the dependency analysis.
+    /// The new mode implements the same functionality as the [TypedMode] and additionally contains the dependencies between streams in the specification.
+    /// The function moves the information of the previous mode to the new one and therefore destroys the current mode.
+    ///
+    /// # Fails
+    /// The function returns a [RtLolaError] if the specification is not well-formed.
+    pub fn analyze_dependencies(self) -> Result<Hir<DepAnaMode>, RtLolaError> {
+        self.progress()
+    }
+}
+
+/// Represents the results of the dependency analysis
+#[derive(Debug, Clone)]
+pub struct DepAna {
+    direct_accesses: Streamdependencies,
+    transitive_accesses: Streamdependencies,
+    direct_accessed_by: Streamdependencies,
+    transitive_accessed_by: Streamdependencies,
+    aggregated_by: Windowdependencies,
+    aggregates: Windowdependencies,
+    graph: DependencyGraph,
+}
+
+/// Represents the mode after the dependency analysis
+///
+/// This struct represents the mode after the dependency analysis.
+/// Besides this result, this mode has the same functionality as all the previous modes.
+/// The [DepAnaTrait] defines the new functionality of the mode.
+#[covers_functionality(TypedTrait, types)]
+#[covers_functionality(DepAnaTrait, dependencies)]
+#[derive(Debug, Clone, HirMode)]
+pub struct DepAnaMode {
+    types: Typed,
+    dependencies: DepAna,
+}
+
+/// Describes the functionality of a mode after analyzing the dependencies
+#[mode_functionality]
+pub trait DepAnaTrait {
+    /// Returns all streams that are direct accessed by `who`
+    ///
+    /// The function returns all streams that are direct accessed by `who`.
+    /// A stream `who` accesses a stream `res`, if the stream expression, the spawn condition and definition, the filter condition, or the close condition of 'who' has a stream or window lookup to `res`.
+    /// Direct accesses are all accesses appearing in the expressions of the stream itself.
+    fn direct_accesses(&self, who: SRef) -> Vec<SRef>;
+
+    /// Returns all streams that are transitive accessed by `who`
+    ///
+    /// The function returns all streams that are transitive accessed by `who`.
+    /// A stream `who` accesses a stream `res`, if the stream expression, the spawn condition and definition, the filter condition, or the close condition of 'who' has a stream or window lookup to 'res'.
+    /// Transitive accesses are all accesses appearing in the expressions of the stream itself or indirect by another stream lookup.
+    fn transitive_accesses(&self, who: SRef) -> Vec<SRef>;
+
+    /// Returns all streams that direct access `who`
+    ///
+    /// The function returns all streams that direct access `who`.
+    /// A stream `who` is accessed by a stream `res`, if the stream expression, the spawn condition and definition, the filter condition, or the close condition of 'res' has a stream or window lookup to 'who'.
+    /// Direct accesses are all accesses appearing in the expressions of the stream itself.
+    fn direct_accessed_by(&self, who: SRef) -> Vec<SRef>;
+
+    /// Returns all streams that transitive access `who`
+    ///
+    /// The function returns all streams that transitive access `who`.
+    /// A stream `who` is accessed by a stream `res`, if the stream expression, the spawn condition and definition, the filter condition, or the close condition of 'res' has a stream or window lookup to 'who'.
+    /// Transitive accesses are all accesses appearing in the expressions of the stream itself or indirect by another stream lookup.
+    fn transitive_accessed_by(&self, who: SRef) -> Vec<SRef>;
+
+    /// Returns all windows that aggregate `who` and the stream that uses the window
+    ///
+    /// The function returns all windows that aggregate `who` and the stream that uses the window.
+    /// The result contains only the windows that are direct.
+    fn aggregated_by(&self, who: SRef) -> Vec<(SRef, WRef)>; // (non-transitive)
+
+    /// Returns all windows that are used in `who` and the corresponding stream that is aggregated
+    ///
+    /// The function returns all windows that are used in `who` and the corresponding stream that is aggregated.
+    /// The result contains only the windows that are direct.
+    fn aggregates(&self, who: SRef) -> Vec<(SRef, WRef)>; // (non-transitive)
+
+    /// Returns the (Dependency Graph)[DependencyGraph] of the specification
+    fn graph(&self) -> &DependencyGraph;
+}
+
+impl HirStage for Hir<DepAnaMode> {
     type NextStage = OrderedMode;
 
-    fn progress(self, _handler: &Handler) -> Result<Hir<Self::NextStage>, Self::Error> {
+    fn progress(self) -> Result<Hir<Self::NextStage>, RtLolaError> {
         let order = Ordered::analyze(&self);
 
         let mode = OrderedMode {
@@ -283,21 +275,20 @@ impl HirStage for Hir<TypedMode> {
     }
 }
 
-impl Hir<TypedMode> {
+impl Hir<DepAnaMode> {
     /// Returns the [RtLolaHir](crate::RtLolaHir) with the spawn and evaluation layer of each stream
     ///
     /// # Fails
     /// The function fails if the evaluation order cannot be determined.
-    pub fn determine_evaluation_order(self, handler: &Handler) -> Result<Hir<OrderedMode>, OrderErr> {
-        self.progress(handler)
+    pub fn determine_evaluation_order(self) -> Result<Hir<OrderedMode>, RtLolaError> {
+        self.progress()
     }
 }
 
 /// Represents the evaluation order
 #[derive(Debug, Clone)]
 pub struct Ordered {
-    event_layers: HashMap<SRef, StreamLayers>,
-    periodic_layers: HashMap<SRef, StreamLayers>,
+    stream_layers: HashMap<SRef, StreamLayers>,
 }
 
 /// Represents the mode after determining the evaluation order
@@ -326,10 +317,9 @@ pub trait OrderedTrait {
 }
 
 impl HirStage for Hir<OrderedMode> {
-    type Error = MemBoundErr;
     type NextStage = MemBoundMode;
 
-    fn progress(self, _handler: &Handler) -> Result<Hir<Self::NextStage>, Self::Error> {
+    fn progress(self) -> Result<Hir<Self::NextStage>, RtLolaError> {
         let memory = MemBound::analyze(&self, false);
 
         let mode = MemBoundMode {
@@ -353,11 +343,11 @@ impl HirStage for Hir<OrderedMode> {
 
 impl Hir<OrderedMode> {
     /// Returns the [RtLolaHir](crate::RtLolaHir) with the memory-bound for each stream
-    ///     
+    ///
     /// # Fails
     /// The function fails if the memory cannot be determined.
-    pub fn determine_memory_bounds(self, handler: &Handler) -> Result<Hir<MemBoundMode>, MemBoundErr> {
-        self.progress(handler)
+    pub fn determine_memory_bounds(self) -> Result<Hir<MemBoundMode>, RtLolaError> {
+        self.progress()
     }
 }
 
@@ -395,10 +385,9 @@ pub trait MemBoundTrait {
 }
 
 impl HirStage for Hir<MemBoundMode> {
-    type Error = CompletionErr;
     type NextStage = CompleteMode;
 
-    fn progress(self, _handler: &Handler) -> Result<Hir<Self::NextStage>, Self::Error> {
+    fn progress(self) -> Result<Hir<Self::NextStage>, RtLolaError> {
         let mode = CompleteMode {
             dependencies: self.mode.dependencies,
             types: self.mode.types,
@@ -418,17 +407,14 @@ impl HirStage for Hir<MemBoundMode> {
     }
 }
 
-#[derive(Debug, Clone, Copy)]
-pub enum CompletionErr {}
-
 impl Hir<MemBoundMode> {
     /// Returns the [RtLolaHir](crate::RtLolaHir) in the last mode
     ///
     /// The function returns the [RtLolaHir](crate::RtLolaHir) in the [CompleteMode].
     /// This mode indicates that the [RtLolaHir](crate::RtLolaHir) has passed all analyzes and now contains all information.
     /// The function moves the information of the previous mode to the new one and therefore destroys the current mode.
-    pub fn finalize(self, handler: &Handler) -> Result<Hir<CompleteMode>, CompletionErr> {
-        self.progress(handler)
+    pub fn finalize(self) -> Result<Hir<CompleteMode>, RtLolaError> {
+        self.progress()
     }
 }
 
