@@ -2,6 +2,7 @@ use std::collections::HashMap;
 
 use petgraph::algo::is_cyclic_directed;
 use petgraph::Outgoing;
+use serde::{Deserialize, Serialize};
 
 use super::{Ordered, OrderedTrait, TypedTrait};
 use crate::hir::{Hir, SRef};
@@ -10,23 +11,16 @@ use crate::modes::{DepAnaTrait, DependencyGraph, HirMode};
 
 impl OrderedTrait for Ordered {
     fn stream_layers(&self, sr: SRef) -> StreamLayers {
-        match self.event_layers.get(&sr) {
-            Some(layer) => *layer,
-            None => self.periodic_layers[&sr],
-        }
+        self.stream_layers[&sr]
     }
 }
 
-/// Represents the error occuring during the evaluation order determination
-#[derive(Debug, Clone, Copy)]
-pub enum OrderErr {}
-
 /// Represents a layer indicating the position when an expression can be evaluated
-#[derive(Debug, Clone, Copy, PartialOrd, Ord, PartialEq, Eq)]
+#[derive(Debug, Clone, Copy, PartialOrd, Ord, PartialEq, Eq, Serialize, Deserialize)]
 pub struct Layer(usize);
 
 /// Wrapper to collect the layer when a stream instance is spawned and evaluated
-#[derive(Debug, Clone, Copy, PartialOrd, Ord, PartialEq, Eq)]
+#[derive(Debug, Clone, Copy, PartialOrd, Ord, PartialEq, Eq, Serialize, Deserialize)]
 pub struct StreamLayers {
     spawn: Layer,
     evaluation: Layer,
@@ -81,13 +75,8 @@ impl Ordered {
         M: HirMode + DepAnaTrait + TypedTrait,
     {
         // split graph in periodic and event-based
-        let (event_graph, periodic_graph) = spec.graph().split_graph(spec);
-        let event_layers = Self::compute_layers(spec, &event_graph, true);
-        let periodic_layers = Self::compute_layers(spec, &periodic_graph, false);
-        Ordered {
-            event_layers,
-            periodic_layers,
-        }
+        let stream_layers = Self::compute_layers(spec, spec.graph());
+        Ordered { stream_layers }
     }
 
     /// Returns the spawn and evaluation layer of for either each event-based stream or periodic stream
@@ -96,43 +85,45 @@ impl Ordered {
     /// The first graph only contains the lookups occurring in the spawn part of the stream template.
     /// The analysis computes from this graph the spawn layers of each stream.
     /// The function computes from the second graph the evaluation layers of each stream.
-    fn compute_layers<M>(spec: &Hir<M>, graph: &DependencyGraph, is_event: bool) -> HashMap<SRef, StreamLayers>
+    fn compute_layers<M>(spec: &Hir<M>, graph: &DependencyGraph) -> HashMap<SRef, StreamLayers>
     where
         M: HirMode + DepAnaTrait + TypedTrait,
     {
         // Prepare graphs
-        let graph = graph.without_negative_offset_edges().without_close_edges();
-        let spawn_graph = graph.only_spawn_edges();
+        let graph = &graph
+            .clone()
+            .without_negative_offset_edges()
+            .without_close()
+            .without_different_pacing(spec);
+        let spawn_graph = &graph.clone().only_spawn();
+
         debug_assert!(
             !is_cyclic_directed(&graph),
             "This should be already checked in the dependency analysis."
         );
 
         // start analysis
-        let mut evaluation_layers = if is_event {
-            spec.inputs()
-                .map(|i| (i.sr, Layer::new(0)))
-                .collect::<HashMap<SRef, Layer>>()
-        } else {
-            HashMap::new()
-        };
-        let mut spawn_layers = if is_event {
-            spec.inputs()
-                .map(|i| (i.sr, Layer::new(0)))
-                .collect::<HashMap<SRef, Layer>>()
-        } else {
-            HashMap::new()
-        };
+        let mut evaluation_layers = spec
+            .inputs()
+            .map(|i| (i.sr, Layer::new(0)))
+            .collect::<HashMap<SRef, Layer>>();
+        let mut spawn_layers = spec
+            .inputs()
+            .map(|i| (i.sr, Layer::new(0)))
+            .collect::<HashMap<SRef, Layer>>();
+
         while graph.node_count() != evaluation_layers.len() {
             // build spawn layers
             spawn_graph.node_indices().for_each(|node| {
                 let sref = spawn_graph.node_weight(node).unwrap();
+                // If we dont know the spawn layer yet
                 if !spawn_layers.contains_key(sref) {
+                    // get evaluation layer of successors
                     let neighbor_layers: Vec<_> = spawn_graph
                         .neighbors_directed(node, Outgoing)
                         .map(|outgoing_neighbor| {
                             evaluation_layers
-                                .get(&spawn_graph.node_weight(outgoing_neighbor).unwrap())
+                                .get(spawn_graph.node_weight(outgoing_neighbor).unwrap())
                                 .copied()
                         })
                         .collect();
@@ -158,31 +149,26 @@ impl Ordered {
             });
             graph.node_indices().for_each(|node| {
                 let sref = graph.node_weight(node).unwrap();
-                // build evaluation layers
+                // if we dont know the evaluation layer, but the spawn layer is known
                 if !evaluation_layers.contains_key(sref) && spawn_layers.contains_key(sref) {
-                    // Layer for current stream check incoming
+                    // Get evaluation layer of successors
                     let neighbor_layers: Vec<_> = graph
                         .neighbors_directed(node, Outgoing)
                         .flat_map(|outgoing_neighbor| {
-                            if outgoing_neighbor == node {
-                                None
-                            } else {
-                                Some(outgoing_neighbor)
-                            }
-                        }) // delete self references
+                            //ignore selfloops
+                            (outgoing_neighbor != node).then(|| outgoing_neighbor)
+                        })
                         .map(|outgoing_neighbor| {
                             evaluation_layers
-                                .get(&graph.node_weight(outgoing_neighbor).unwrap())
+                                .get(graph.node_weight(outgoing_neighbor).unwrap())
                                 .copied()
                         })
                         .collect();
                     let computed_evaluation_layer = if neighbor_layers.is_empty() {
-                        if is_event {
-                            Some(Layer::new(1))
-                        } else {
-                            Some(Layer::new(0))
-                        }
+                        // There are no successors
+                        Some(Layer::new(1))
                     } else {
+                        // eval_layer = max(successor_eval_layers) + 1
                         neighbor_layers
                             .into_iter()
                             .fold(Some(Layer::new(0)), |cur_res_layer, neighbor_layer| {
@@ -196,6 +182,7 @@ impl Ordered {
                             .map(|layer| Layer::new(layer.inner() + 1))
                     };
                     if let Some(layer) = computed_evaluation_layer {
+                        // Evaluation layer has to be greater than spawn layer
                         let layer = if spawn_layers[sref] < layer {
                             layer
                         } else {
@@ -215,41 +202,23 @@ impl Ordered {
 
 #[cfg(test)]
 mod tests {
-    use std::path::PathBuf;
-
-    use rtlola_parser::{parse_with_handler, ParserConfig};
-    use rtlola_reporting::Handler;
+    use rtlola_parser::{parse, ParserConfig};
 
     use super::*;
     use crate::modes::BaseMode;
-    fn check_eval_order_for_spec(
-        spec: &str,
-        ref_event_layers: HashMap<SRef, StreamLayers>,
-        ref_periodic_layers: HashMap<SRef, StreamLayers>,
-    ) {
-        let handler = Handler::new(PathBuf::new(), spec.into());
-        let ast = parse_with_handler(ParserConfig::for_string(spec.to_string()), &handler)
-            .unwrap_or_else(|e| panic!("{}", e));
-        let hir = Hir::<BaseMode>::from_ast(ast, &handler)
+    fn check_eval_order_for_spec(spec: &str, ref_layers: HashMap<SRef, StreamLayers>) {
+        let ast = parse(ParserConfig::for_string(spec.to_string())).unwrap_or_else(|e| panic!("{:?}", e));
+        let hir = Hir::<BaseMode>::from_ast(ast)
             .unwrap()
-            .analyze_dependencies(&handler)
+            .check_types()
             .unwrap()
-            .check_types(&handler)
+            .analyze_dependencies()
             .unwrap();
         let order = Ordered::analyze(&hir);
-        let Ordered {
-            event_layers,
-            periodic_layers,
-        } = order;
-        assert_eq!(event_layers.len(), ref_event_layers.len());
-        event_layers.iter().for_each(|(sr, layers)| {
-            let ref_layers = &ref_event_layers[sr];
-            assert_eq!(ref_layers.spawn_layer(), layers.spawn_layer());
-            assert_eq!(ref_layers.evaluation_layer(), layers.evaluation_layer());
-        });
-        assert_eq!(periodic_layers.len(), ref_periodic_layers.len());
-        periodic_layers.iter().for_each(|(sr, layers)| {
-            let ref_layers = &ref_periodic_layers[sr];
+        let Ordered { stream_layers } = order;
+        assert_eq!(stream_layers.len(), ref_layers.len());
+        stream_layers.iter().for_each(|(sr, layers)| {
+            let ref_layers = &ref_layers[sr];
             assert_eq!(ref_layers.spawn_layer(), layers.spawn_layer());
             assert_eq!(ref_layers.evaluation_layer(), layers.evaluation_layer());
         });
@@ -268,8 +237,7 @@ mod tests {
         ]
         .into_iter()
         .collect();
-        let periodic_layers = HashMap::new();
-        check_eval_order_for_spec(spec, event_layers, periodic_layers)
+        check_eval_order_for_spec(spec, event_layers)
     }
 
     #[test]
@@ -292,8 +260,7 @@ mod tests {
         ]
         .into_iter()
         .collect();
-        let periodic_layers = HashMap::new();
-        check_eval_order_for_spec(spec, event_layers, periodic_layers)
+        check_eval_order_for_spec(spec, event_layers)
     }
 
     #[test]
@@ -309,8 +276,7 @@ mod tests {
         ]
         .into_iter()
         .collect();
-        let periodic_layers = HashMap::new();
-        check_eval_order_for_spec(spec, event_layers, periodic_layers)
+        check_eval_order_for_spec(spec, event_layers)
     }
 
     #[test]
@@ -326,13 +292,9 @@ mod tests {
         ]
         .into_iter()
         .collect::<HashMap<&str, SRef>>();
-        let event_layers = vec![
+        let ref_layers = vec![
             (sname_to_sref["a"], StreamLayers::new(Layer::new(0), Layer::new(0))),
             (sname_to_sref["c"], StreamLayers::new(Layer::new(0), Layer::new(1))),
-        ]
-        .into_iter()
-        .collect();
-        let periodic_layers = vec![
             (sname_to_sref["b"], StreamLayers::new(Layer::new(0), Layer::new(1))),
             (sname_to_sref["d"], StreamLayers::new(Layer::new(0), Layer::new(1))),
             (sname_to_sref["e"], StreamLayers::new(Layer::new(0), Layer::new(2))),
@@ -340,7 +302,7 @@ mod tests {
         ]
         .into_iter()
         .collect();
-        check_eval_order_for_spec(spec, event_layers, periodic_layers)
+        check_eval_order_for_spec(spec, ref_layers)
     }
 
     #[test]
@@ -354,19 +316,15 @@ mod tests {
         ]
         .into_iter()
         .collect::<HashMap<&str, SRef>>();
-        let event_layers = vec![
+        let ref_layers = vec![
             (sname_to_sref["a"], StreamLayers::new(Layer::new(0), Layer::new(0))),
             (sname_to_sref["c"], StreamLayers::new(Layer::new(0), Layer::new(1))),
-        ]
-        .into_iter()
-        .collect();
-        let periodic_layers = vec![
             (sname_to_sref["b"], StreamLayers::new(Layer::new(0), Layer::new(1))),
             (sname_to_sref["d"], StreamLayers::new(Layer::new(0), Layer::new(1))),
         ]
         .into_iter()
         .collect();
-        check_eval_order_for_spec(spec, event_layers, periodic_layers)
+        check_eval_order_for_spec(spec, ref_layers)
     }
 
     #[test]
@@ -390,8 +348,7 @@ mod tests {
         ]
         .into_iter()
         .collect();
-        let periodic_layers = HashMap::new();
-        check_eval_order_for_spec(spec, event_layers, periodic_layers)
+        check_eval_order_for_spec(spec, event_layers)
     }
     #[test]
     fn negative_loop_different_offsets() {
@@ -412,8 +369,7 @@ mod tests {
         ]
         .into_iter()
         .collect();
-        let periodic_layers = HashMap::new();
-        check_eval_order_for_spec(spec, event_layers, periodic_layers)
+        check_eval_order_for_spec(spec, event_layers)
     }
 
     #[test]
@@ -435,8 +391,7 @@ mod tests {
         ]
         .into_iter()
         .collect();
-        let periodic_layers = HashMap::new();
-        check_eval_order_for_spec(spec, event_layers, periodic_layers)
+        check_eval_order_for_spec(spec, event_layers)
     }
 
     #[test]
@@ -466,8 +421,7 @@ mod tests {
         ]
         .into_iter()
         .collect();
-        let periodic_layers = HashMap::new();
-        check_eval_order_for_spec(spec, event_layers, periodic_layers)
+        check_eval_order_for_spec(spec, event_layers)
     }
 
     #[test]
@@ -483,20 +437,16 @@ mod tests {
         ]
         .into_iter()
         .collect::<HashMap<&str, SRef>>();
-        let event_layers = vec![
+        let ref_layers = vec![
             (sname_to_sref["a"], StreamLayers::new(Layer::new(0), Layer::new(0))),
             (sname_to_sref["b"], StreamLayers::new(Layer::new(0), Layer::new(0))),
             (sname_to_sref["e"], StreamLayers::new(Layer::new(0), Layer::new(1))),
-        ]
-        .into_iter()
-        .collect();
-        let periodic_layers = vec![
             (sname_to_sref["c"], StreamLayers::new(Layer::new(0), Layer::new(1))),
             (sname_to_sref["d"], StreamLayers::new(Layer::new(0), Layer::new(2))),
         ]
         .into_iter()
         .collect();
-        check_eval_order_for_spec(spec, event_layers, periodic_layers)
+        check_eval_order_for_spec(spec, ref_layers)
     }
 
     #[test]
@@ -518,8 +468,7 @@ mod tests {
         ]
         .into_iter()
         .collect();
-        let periodic_layers = vec![].into_iter().collect();
-        check_eval_order_for_spec(spec, event_layers, periodic_layers)
+        check_eval_order_for_spec(spec, event_layers)
     }
     #[test]
     fn sliding_windows_chain_and_hold_lookup() {
@@ -532,17 +481,15 @@ mod tests {
         ]
         .into_iter()
         .collect::<HashMap<&str, SRef>>();
-        let event_layers = vec![(sname_to_sref["a"], StreamLayers::new(Layer::new(0), Layer::new(0)))]
-            .into_iter()
-            .collect();
-        let periodic_layers = vec![
+        let ref_layers = vec![
             (sname_to_sref["b"], StreamLayers::new(Layer::new(0), Layer::new(1))),
             (sname_to_sref["c"], StreamLayers::new(Layer::new(0), Layer::new(2))),
             (sname_to_sref["d"], StreamLayers::new(Layer::new(0), Layer::new(2))),
+            (sname_to_sref["a"], StreamLayers::new(Layer::new(0), Layer::new(0))),
         ]
         .into_iter()
         .collect();
-        check_eval_order_for_spec(spec, event_layers, periodic_layers)
+        check_eval_order_for_spec(spec, ref_layers)
     }
 
     #[test]
@@ -558,13 +505,12 @@ mod tests {
         ]
         .into_iter()
         .collect();
-        let periodic_layers = vec![].into_iter().collect();
-        check_eval_order_for_spec(spec, event_layers, periodic_layers)
+        check_eval_order_for_spec(spec, event_layers)
     }
 
     #[test]
     fn lookup_chain_with_parametrization() {
-        let spec = "input a: Int8\noutput b(para) spawn with a if a > 6 := a + para\noutput c(para) spawn with a if a > 6 := a + b(a)\noutput d(para) spawn with a if a > 6 := a + c(a)";
+        let spec = "input a: Int8\noutput b(para) spawn with a if a > 6 := a + para\noutput c(para) spawn with a if a > 6 := a + b(para)\noutput d(para) spawn with a if a > 6 := a + c(para)";
         let sname_to_sref = vec![
             ("a", SRef::In(0)),
             ("b", SRef::Out(0)),
@@ -581,13 +527,12 @@ mod tests {
         ]
         .into_iter()
         .collect();
-        let periodic_layers = vec![].into_iter().collect();
-        check_eval_order_for_spec(spec, event_layers, periodic_layers)
+        check_eval_order_for_spec(spec, event_layers)
     }
 
     #[test]
     fn parameter_loop_with_lookup_in_close() {
-        let spec = "input a: Int8\ninput b: Int8\noutput c(p) spawn with a if a < b := p + b + g(p).hold().defaults(to: 0)\noutput d(p) spawn with b if c(4).hold().defaults(to: 0) < 4 := b + 5\noutput e(p) spawn with b := d(p).hold().defaults(to: 0) + b\noutput f(p) spawn with b filter e(p).hold().defaults(to: 0) < 6 := b + 5\noutput g(p) spawn with b close f(p).hold().defaults(to: 0) < 6 := b + 5";
+        let spec = "input a: Int8\ninput b: Int8\noutput c(p) spawn with a if a < b := p + b + g(p).hold().defaults(to: 0)\noutput d(p) spawn with b if c(4).hold().defaults(to: 0) < 4 := b + 5\noutput e(p) spawn with b := d(p).hold().defaults(to: 0) + b\noutput f(p) spawn with b filter e(p).hold().defaults(to: 0) < 6 := b + 5\noutput g(p) spawn with b close @true f(p).hold().defaults(to: 0) < 6 := b + 5";
         let sname_to_sref = vec![
             ("a", SRef::In(0)),
             ("b", SRef::In(1)),
@@ -610,8 +555,7 @@ mod tests {
         ]
         .into_iter()
         .collect();
-        let periodic_layers = vec![].into_iter().collect();
-        check_eval_order_for_spec(spec, event_layers, periodic_layers)
+        check_eval_order_for_spec(spec, event_layers)
     }
 
     #[test]
@@ -633,8 +577,7 @@ mod tests {
         ]
         .into_iter()
         .collect();
-        let periodic_layers = vec![].into_iter().collect();
-        check_eval_order_for_spec(spec, event_layers, periodic_layers)
+        check_eval_order_for_spec(spec, event_layers)
     }
 
     #[test]
@@ -658,7 +601,43 @@ mod tests {
         ]
         .into_iter()
         .collect();
-        let periodic_layers = vec![].into_iter().collect();
-        check_eval_order_for_spec(spec, event_layers, periodic_layers)
+        check_eval_order_for_spec(spec, event_layers)
+    }
+
+    #[test]
+    fn test_spawn_eventbased() {
+        let spec = "input a: Int32\n\
+                  output b(x: Int32) spawn with a := x + a";
+        let sname_to_sref = vec![("a", SRef::In(0)), ("b", SRef::Out(0))]
+            .into_iter()
+            .collect::<HashMap<&str, SRef>>();
+        let event_layers = vec![
+            (sname_to_sref["a"], StreamLayers::new(Layer::new(0), Layer::new(0))),
+            (sname_to_sref["b"], StreamLayers::new(Layer::new(1), Layer::new(2))),
+        ]
+        .into_iter()
+        .collect();
+        check_eval_order_for_spec(spec, event_layers)
+    }
+
+    #[test]
+    fn test_delay() {
+        let spec = "input a: UInt64\n\
+                            output a_counter: UInt64 @a := a_counter.offset(by: -1).defaults(to: 0) + 1\n\
+                            output b(p: UInt64) @1Hz spawn with a_counter if a = 1 close if true then true else b(p) := a.hold(or: 0) == 2";
+        let sname_to_sref = vec![("a", SRef::In(0)), ("a_counter", SRef::Out(0)), ("b", SRef::Out(1))]
+            .into_iter()
+            .collect::<HashMap<&str, SRef>>();
+        let ref_layers = vec![
+            (sname_to_sref["a"], StreamLayers::new(Layer::new(0), Layer::new(0))),
+            (
+                sname_to_sref["a_counter"],
+                StreamLayers::new(Layer::new(0), Layer::new(1)),
+            ),
+            (sname_to_sref["b"], StreamLayers::new(Layer::new(2), Layer::new(3))),
+        ]
+        .into_iter()
+        .collect();
+        check_eval_order_for_spec(spec, ref_layers)
     }
 }
